#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <dmmlib/heap.h>
#include "dmm_config.h"
#ifdef HAVE_LOCKS
#include "posix_lock.h"
#endif /* HAVE_LOCKS */
#include "other.h"
#include "sys_alloc.h"
#include "block_header.h"

#ifndef WITH_MEMORY_SPACE_AWARENESS
void *sys_alloc(heap_t *heap, size_t size) {
#else
void *sys_alloc(allocator_t *allocator, heap_t *heap, size_t size) {
#endif /* WITH_MEMORY_SPACE_AWARENESS */

    size_t allocation_size;
    void *ptr;

#ifdef HAVE_LOCKS
    sbrk_lock();
#endif /* HAVE_LOCKS */

    allocation_size = req_padding(size) + HEADER_SIZE;

#ifndef WITH_MEMORY_SPACE_AWARENESS
    ptr = sbrk((int) allocation_size);
    if(ptr == (void *) -1) {
        printf("sbrk problem for size of: %zu\n", allocation_size);
        printf( "Error on sbrk: %s\n", strerror( errno ) );
    }
#else
    if(allocator->remaining_size >= allocation_size) {
        ptr = allocator->border_ptr;
        allocator->border_ptr = (char *) allocator->border_ptr + allocation_size;
        allocator->remaining_size -= allocation_size;
    } else {
        printf("No more free space.\n");
    }
#endif /* WITH_MEMORY_SPACE_AWARENESS */

    ptr = (void *) ((char *) ptr + HEADER_SIZE);

    set_size(ptr, req_padding(size));
    heap->dmm_stats.mem_allocated += req_padding(size);
    heap->dmm_stats.mem_requested += size;

#ifdef HAVE_LOCKS
    sbrk_unlock();
#endif /* HAVE_LOCKS */

    return  ptr;
}

